/* lib.c  09-Apr-89  A.J.Travis */

/*
 * Part of the 'standard' C library
 */

/* JGH: Increase LINESIZE from 81 to 256 */

#include <stdio.h>

#define _MAXARG 32			/* max no. of command arguments */
#define _IOSTRG -1			/* string input mode for _doscan() */
#define LINESIZE 256			/* length of output line + 1 */
#define EOL '\r'			/* MOS end of line */
#define DELETE '\177'			/* i/o delete char */
#define SAME 0				/* strcmp() strings equal */

#define TRUE 1
#define FALSE 0
#define ERROR -1

/*
 * append string s2 to string s1
 */
strcat(s1, s2)
char *s1;
char *s2;
{
	char *os1;

	os1 = s1;
	while (*s1++)
		;
	--s1;
	while (*s1++ = *s2++)
		;
	return(os1);
}

/*
 * compare strings s1 and s2
 */
strcmp(s1, s2)
char *s1;
char *s2;
{
	while (*s1 == *s2++) {
		if (*s1++ == '\0')
			return(0);
	}
	return(*s1 - *--s2);
}

/*
 * copy string s2 to string s1
 */
strcpy(s1, s2)
char *s1;
char *s2;
{
	char *os1;

	os1 = s1;
	while (*s1++ = *s2++)
		;
	return(os1);
}

/*
 * length of string
 */
strlen(s)
char *s;
{
	int n;

	n = 0;
	while (*s++)
		n++;
	return(n);
}

/*
 * concatenate up to n chars of s2 onto the end of s1
 */
strncat(s1, s2, n)
char *s1;
char *s2;
int n;
{
	char *os1;

	os1 = s1;
	while (*s1++)
		;
	--s1;
	while  (*s1++ = *s2++)
		if (--n < 0) {
			*--s1 = '\0';
			break;
		}
	return(os1);
}

/*
 * compare up to n chars of s1 and s2
 */
strncmp(s1, s2, n)
char *s1;
char *s2;
int n;
{
	while (--n >= 0 & *s1 == *s2++)
		if (*s1++ == '\0')
			return(0);
	if (n < 0)
		return(0);
	else
		return(*s1 - *--s2);
}

/*
 * copy n chars from s2 to s1, truncating or null padding if necessary
 */
strncpy(s1, s2, n)
char *s1;
char *s2;
int n;
{
	int i;
	char *os1;

	os1 = s1;
	for (i = 0; i < n; i++)
		if ((*s1++ = *s2++) == '\0') {
			while (++i < n)
				*s1++ = '\0';
			return(os1);
		}
	return(os1);
}

/*
 * close i/o stream
 */
fclose(fp)
FILE *fp;		/* (pseudo) i/o stream pointer */
{
	return(close(fp));	
}

/*
 * open file by pathname
 */
fopen(name, mode)
char *name;
char *mode;
{
	int fd;			/* file descriptor for low level i/o */

	if (*mode == 'w') {

		/* need to zap existing file under DFS/ADFS */
		if ((fd = open(name, 2)) != -1) {
			close(fd);
			unlink(name);
		}

		/* create file (pmode ignored in this version ...) */
		fd = creat(name, 0);
	}
	else
		/* anything else is assumed to be 'r' */
		fd = open(name, 0);
	if (fd == -1)
		return(NULL);

	/* no file pointers in Small-C */
	return(fd);
}

/*
 * a version of fgets that permits '\r' in lieu of
 * '\n' as end of line and allows character deletion
 */
fgets(s, n, fp)
char *s;
int n;
FILE *fp;
{
	int c;
	char *cs;

	cs = s;
	while (--n > 0 & (c = getc(fp)) != EOF) {
		if (c == DELETE) {
			if (--cs < s)
				cs = s;
			continue;
		}
		if (c == '\r')
			c = '\n';
		if ((*cs++ = c) == '\n')
			break;
	}
	*cs = '\0';
	if (c == EOF & cs == s)
		return(NULL);
	else
		return(s);
}

/*
 * Put null terminated string to output stream
 * -------------------------------------------
 * map '\n' to '\r' for BBC Micro environment
 */
fputs(s, fp)
char *s;
FILE *fp;
{
	char *p;
	int r;
	int c;

	if ((fp != stdin) & (fp != stdout) & (fp != stderr)) {
		for (p = s; *p; ++p) {
			if (*p == '\n')
				*p = '\r';
		}
	}
	while (c = *s++)
		r = putc(c, fp);
	return(r);
}

gets(s)
char *s;
{
	int c;
	char *cs;

	cs = s;
	while ((c = getchar()) != '\n' & c >= 0)
		*cs++ = c;
	if (c < 0 & cs == s)
		return(NULL);
	else {
		*cs = '\0';
		return(s);
	}
}

puts(s)
char *s;
{
	int c;

	while (c = *s++)
		putchar(c);
	return(putchar('\n'));
}

/*
 * formatted output
 */
printf(fmt, arg)
char *fmt;	/* output format */
int arg;	/* argument list */
{
	char _iobuf[LINESIZE];
	
	_doprint(_iobuf, fmt, &arg);
	fputs(_iobuf, stdout);
}

/*
 * formatted output to file
 */
fprintf(fp, fmt, arg)
FILE *fp;	/* output stream */
char *fmt;	/* output format */
int arg;	/* argument list */
{
 	char _iobuf[LINESIZE];
	
	_doprint(_iobuf, fmt, &arg);
	fputs(_iobuf, fp);
}

/*
 * formatted output to string
 */
sprintf(s, fmt, arg)
char *s;	/* output string */
char *fmt;	/* output format */
int arg;	/* argument list */
{
	_doprint(s, fmt, &arg);
}

/*
 * output format conversion
 */
_doprint(buf, fmt, arg)
char *buf;	/* output buffer */
char *fmt;	/* output format */
int *arg[];	/* argument list */
{
	char *cp;	/* pointer to start of conversion buffer */
	char *bp;	/* buffer pointer */
	char pad;	/* padding character */
	char c;		/* conversion character */
	char *s;	/* string pointer */
	int width;	/* minimum field width */
	int precision;	/* number of digits to print */

	bp = buf;
	while (*fmt) {
		cp = bp;
		if (*fmt != '%') {
			*bp++ = *fmt++;
			continue;
		}
		if  (*++fmt == '-')
			pad = *fmt++;
		else
			pad = ' ';
		if (*fmt == '0') {
			if (pad == '-')
				*fmt++;
			else
				pad = *fmt++;
		}
		width = 0;
		while (isdigit(*fmt))
			width = width * 10 + *fmt++ - '0';
		if (*fmt != '.')
			precision = -1;
		else {
			if (isdigit(*++fmt) == 0)
				precision = 1;
			else {
				precision = 0;
				while (isdigit(*fmt))
					precision = precision * 10 + *fmt++ - '0';
			}
		}
		if ((c = *fmt++) == 'd') {
			if (*arg & 0x8000) {
				*bp++ = '-';
				bp = bp + itoa(-*arg, bp, 10);
			}
			else
				bp = bp + itoa(*arg, bp, 10);
		}
		else if (c == 'o')
			bp = bp + itoa(*arg, bp, 8);
		else if (c == 'u')
			bp = bp + itoa(*arg, bp, 10);
		else if (c == 'x')
			bp = bp + itoa(*arg, bp, 16);
		else if (c == 'X')
			bp = bp + itoa(*arg, bp, -16);
		else if (c == 'c') {
			*bp++ = *arg;
			*bp = '\0';
		}
		else if (c == 's') {
			s = *arg;
			--bp;
			if (precision < 0)
				while (*++bp = *s++)
					;
			else {
				while (*++bp = *s++) {
					if (precision-- == 0) {
						*bp = '\0';
						break;
					}
				}
			}
		}
		else {
			*bp++ = c;
			*bp = '\0';
		}
		if ((width = width - (bp - cp)) > 0) {
			if (pad == '-') {
				while (width-- > 0)
					*bp++ = ' ';
				*bp = '\0';
			}
			else {
				if ((c == 'd') & (*cp == '-') & (pad == '0'))
					cp++;
				while (bp >= cp)
					*(bp + width) = *bp--;
				while (width-- > 0)
					*cp++ = pad;
				while (*bp++)
					;
				--bp;
			}
		}
		arg++;
	}
	*bp = '\0';
}

/*
 * convert unsigned 16-bit number to ASCII string
 * ----------------------------------------------
 * returns number of digits in output string
 */
itoa(n, s, base)
unsigned n;	/* input value */
char *s;	/* output buffer */
int base;	/* conversion base */
{
	int d;		/* current digit */
	int xset;	/* extended character set */
	char *p;	/* buffer pointer */

	if (base >= 0)
		/* default is lower-case hex */
		xset = 'a' - 10;
	else {
		/* 'negative' base used to force upper-case hex */
		base = -base;
		xset = 'A' - 10;
	}

	/* generate string least significant digit first */
	p = s;
	if (n == 0)
		*p++ = '0';
	else {
		while (n) {
			if ((d = n % base) < 10)
				*p++ = d + '0';
			else
				*p++ = d + xset;
			n = n / base;
		}
	}
	n = p - s;

	/* reverse digit string in-situ */
	*p-- = '\0';
	while (s < p) {
		d = *s;
		*s++ = *p;
		*p-- = d;
	}
	return(n);
}

/*
 * formatted input
 */
scanf(fmt, arg)
char *fmt;
int arg;
{
	char _iobuf[LINESIZE];

	return(_doscan(stdin, _iobuf, fmt, &arg));
}

/*
 * formatted input from file
 */
fscanf(fp, fmt, arg)
FILE *fp;
char *fmt;
int arg;
{
	char _iobuf[LINESIZE];

	return(_doscan(fp, _iobuf, fmt, &arg));
}

/*
 * formatted input from string
 */
sscanf(s, fmt, arg)
char *s;
char *fmt;
int arg;
{
	return(_doscan(_IOSTRG, s, fmt, &arg));
}

/*
 * input format conversion
 */
_doscan(fp, buf, fmt, arg)
FILE *fp;			/* input stream */
char *buf;			/* i/o buffer */
char *fmt;			/* conversion format */
int *arg[];			/* argument list */
{
	char *bp;		/* conversion buffer pointer */
	char *ip;		/* i/o buffer pointer */
	char fc;		/* current format string character */
	char ic;		/* input character */
	int width;		/* input field width */
	int sign;		/* sign of decimal number */
	int n;			/* number input */
	int digit;		/* hex digit */
	int assign;		/* if using '*' suppression character */
	int nmatch;		/* number of successful conversions */

	/* start with empty buffer if reading from file */
	if (fp != _IOSTRG)
		*buf = '\0';

	ip = buf;
	nmatch = 0;
	while (*fmt) {
		while ((fc != '%') | (ic == '\0')) {
			while (isspace(fc = *fmt++))
				;
			while (isspace(ic = *ip++))
				;
			if (fc == '\0')
				break;

			/* fill buffer unless input from string */
			if (fp != _IOSTRG) {
				while (ic == '\0') {
					ip = buf;
					for (n = 0; n < LINESIZE; n++) {
						if ((ic = getc(fp)) == EOF)
							break;
						if (ic == DELETE) {
							if (--ip < buf)
								ip = buf;
							continue;
						}
						*ip++ = ic;
						if ((ic == EOL) | (ic == '\n'))
							break;
					}
					*ip = '\0';
					if ((ic == EOF) & (ip == buf))
						return(EOF);
					ip = buf;
					while (isspace(ic = *ip++))
						;
				}
			}

			if ((fc != '\0') & (fc != ic) & (fc != '%'))
				return(EOF);
		}
		if (fc == '\0')
			break;
		fc = *fmt++;
		if ((assign = (fc != '*')) == FALSE)
			fc = *fmt++;
		if (isdigit(fc) == 0)
			width = LINESIZE;
		else {
			width = 0;
			do {
				width = (width * 10 + fc - '0');
			}
			while (isdigit(fc = *fmt++));
		}
		n = 0;
		if (fc == 'd') {
			sign = 1;
			if (ic == '+') {
				ic = *ip++;
				--width;
			}
			else if (ic == '-') {
				ic = *ip++;
				sign = -1;
				--width;
			}
			if (isdigit(ic) == 0)
				return(EOF);
			while ((isdigit(ic) != 0) & (width-- > 0)) {
				n = n * 10 + ic - '0';
				ic = *ip++;
			}
			n = n * sign;
		}
		else if (fc == 'o') {
			if ((ic < '0') | (ic > '7'))
				return(EOF);
			while ((ic >= '0') & (ic <= '7') & (width-- > 0)) {
				n = n * 8 + ic - '0';
				ic = *ip++;
			}
		}
		else if (fc == 'x' | fc == 'X') {
			if (ic == '0') {
				if ((ic = *ip++) == fc)
					ic = *ip++;
				else
					return(EOF);
			}
			if (isxdigit(ic) == 0)
				return(EOF);
			while ((isxdigit(ic) != 0) & (width-- > 0)) {
				if (isdigit(ic))
					digit = ic - '0';
				else if (isupper(ic))
					digit = ic - 'A' + 10;
				else if (islower(ic))
					digit = ic - 'a' + 10;
				n = n * 16 + digit;
				ic = *ip++;
			}
		}
		else if (fc == 's') {
			if (assign) {
				bp = arg[nmatch++];
				while ((ic != '\0') & (width-- > 0)) {
					*bp++ = ic;
					if (isspace(ic = *ip++))
						break;
				}
				*bp = '\0';
			}
			else {
				while ((ic != '\0') & (width-- > 0))
					if (isspace(ic = *ip++))
						break;
			}				
		}
		else if (fc == 'c') {
			n = ic;
			ic = *ip++;
			--width;
		}
		else
			return(EOF);
		if ((assign != 0) & (fc != 's'))
			*(arg[nmatch++]) = n;
		--ip;
	}
	return(nmatch);
}

/*
 * convert string to integer
 */
atoi(s)
char *s;
{
	int n;
	int sign;
	int c;

	n = 0;
	sign = 1;			/* no sign == positive */
	if ((c = *s++) == '+')
		c = *s++;
	else if (c == '-') {
		c = *s++;
		sign = -1;
	}
	while (isdigit(c)) {
		n = n * 10 + c - '0';
		c = *s++;
	}
	return(n * sign);
}

/*
 * extract command line arguments
 */
_cmdinit(enter)
unsigned enter;				/* entry point (should be int *()) */
{
	char *lp;			/* pointer to command line */
	char *bp;			/* buffer pointer */
	int argc;			/* argument count */
	int *argv[_MAXARG];		/* BUG - should be char *argv[32] */
	char _argbuf[LINESIZE];		/* argument buffer */

	lp = _cmdline();		/* get address of MOS command line */
	bp = _argbuf;
	argv[0] = bp;			/* argv[0] is name of invocation */
	*bp++ = '\0';			/* argv[0] not accessible to MOS */
	argc = 1;			/* null argv[0] 'always' present */
	while (*lp != EOL) {
		while (*lp == ' ' | *lp == '\t')
			lp++;
		if (*lp != EOL)
			argv[argc++] = bp;
		while (*lp != ' ' & *lp != '\t' & *lp != EOL)
			*bp++ = *lp++;
		*bp++ = '\0';
	}
	enter(argc, argv);
}
